@@ -4,7 +4,7 @@
  */
 
 /*
- * Copyright (C) 2006-2022 Oracle and/or its affiliates.
+ * Copyright (C) 2006-2025 Oracle and/or its affiliates.
  *
  * This file is part of VirtualBox base platform packages, as
  * available from https://www.virtualbox.org.
@@ -54,6 +54,7 @@
 #include <VBox/err.h>
 #include <iprt/mem.h>
 #include <VBox/log.h>
+#include <VBox/VBoxLnxModInline.h>
 #include <iprt/mp.h>
 
 /** @todo figure out the exact version number */
@@ -72,9 +73,18 @@
 # include <asm/msr.h>
 #endif
 
-#include <asm/desc.h>
+#if defined(RT_ARCH_AMD64) || defined(RT_ARCH_X86)
+# include <asm/desc.h>
 
-#include <iprt/asm-amd64-x86.h>
+# include <iprt/asm-amd64-x86.h>
+#endif
+
+#if (RTLNX_VER_RANGE(6,16,0, 6,18,0)) && defined(CONFIG_KVM_GENERIC_HARDWARE_ENABLING) && defined(VBOX_WITH_HOST_VMX)
+# if defined(RT_ARCH_AMD64) || defined(RT_ARCH_X86)
+#  include <linux/kvm_host.h>
+#  define SUPDRV_LINUX_HAS_KVM_HWVIRT_API
+# endif
+#endif
 
 
 /*********************************************************************************************************************************
@@ -109,6 +119,13 @@
 # define VBOX_EXTRA_VERSION_STRING ""
 #endif
 
+/* Macro was renamed in 6.16. */
+# if RTLNX_VER_MIN(6,16,0)
+#  define VBOX_RDMSR_SAFE rdmsrq_safe
+# else
+#  define VBOX_RDMSR_SAFE rdmsrl_safe
+# endif
+
 
 /*********************************************************************************************************************************
 *   Structures and Typedefs                                                                                                      *
@@ -164,6 +181,9 @@ static int  supdrvLinuxLdrModuleNotifyCallback(struct notifier_block *pBlock,
  * Device extention & session data association structure.
  */
 static SUPDRVDEVEXT         g_DevExt;
+#ifdef SUPDRV_LINUX_HAS_KVM_HWVIRT_API
+static bool                 g_fUsingKvmHwvirtApi;
+#endif
 
 /** Module parameter.
  * Not prefixed because the name is used by macros and the end of this file. */
@@ -174,37 +194,6 @@ static int force_async_tsc = 0;
 /** The user device name. */
 #define DEVICE_NAME_USR     "vboxdrvu"
 
-#if (defined(RT_ARCH_AMD64) && RTLNX_VER_MAX(2,6,23)) || defined(VBOX_WITH_TEXT_MODMEM_HACK)
-/**
- * Memory for the executable memory heap (in IPRT).
- */
-# ifdef DEBUG
-#  define EXEC_MEMORY_SIZE   10485760   /* 10 MB */
-# else
-#  define EXEC_MEMORY_SIZE   8388608    /* 8 MB */
-# endif
-extern uint8_t g_abExecMemory[EXEC_MEMORY_SIZE];
-# ifndef VBOX_WITH_TEXT_MODMEM_HACK
-__asm__(".section execmemory, \"awx\", @progbits\n\t"
-        ".align 32\n\t"
-        ".globl g_abExecMemory\n"
-        "g_abExecMemory:\n\t"
-        ".zero " RT_XSTR(EXEC_MEMORY_SIZE) "\n\t"
-        ".type g_abExecMemory, @object\n\t"
-        ".size g_abExecMemory, " RT_XSTR(EXEC_MEMORY_SIZE) "\n\t"
-        ".text\n\t");
-# else
-__asm__(".text\n\t"
-        ".align 4096\n\t"
-        ".globl g_abExecMemory\n"
-        "g_abExecMemory:\n\t"
-        ".zero " RT_XSTR(EXEC_MEMORY_SIZE) "\n\t"
-        ".type g_abExecMemory, @object\n\t"
-        ".size g_abExecMemory, " RT_XSTR(EXEC_MEMORY_SIZE) "\n\t"
-        ".text\n\t");
-# endif
-#endif
-
 /** The file_operations structure. */
 static struct file_operations gFileOpsVBoxDrvSys =
 {
@@ -338,12 +327,12 @@ DECLINLINE(RTGID) vboxdrvLinuxKernGid(void)
 
 
 #ifdef VBOX_WITH_HARDENING
-/** Get the effective UID within the current user namespace. */
-DECLINLINE(RTUID) vboxdrvLinuxEuidInNs(void)
+/** Get the effective UID of the current process. */
+DECLINLINE(RTUID) vboxdrvLinuxEuid(void)
 {
 # if RTLNX_VER_MIN(2,6,29)
 #  if RTLNX_VER_MIN(3,5,0)
-    return from_kuid(current_user_ns(), current->cred->euid);
+    return __kuid_val(current->cred->euid);
 #  else
     return current->cred->euid;
 #  endif
@@ -354,6 +343,35 @@ DECLINLINE(RTUID) vboxdrvLinuxEuidInNs(void)
 #endif
 
 
+#ifdef SUPDRV_LINUX_HAS_KVM_HWVIRT_API
+/**
+ * This is a hack/workaround that attempts to detect whether the kvm_intel/kvm_amd
+ * module is likely to be loaded. See @bugref{10963#c14} for details.
+ */
+static bool is_kvm_hwvirt_mod_likely_loaded(void)
+{
+    /*
+     * We don't disable preemption/interrupts here because we assume all CPUs will have
+     * the relevant bits identical in CR4 and EFER.
+     */
+    bool const     fIsIntel  = ASMIsIntelOrCompatibleCpu();
+    uint64_t const fVmxeMask = RT_BIT_64(13);
+    if (   fIsIntel
+        && (ASMGetCR4() & fVmxeMask))
+        return true;
+
+    bool const     fIsAmd    = ASMIsAmdOrCompatibleCpu();
+    uint32_t const idEfer    = 0xc0000080;
+    uint64_t const fSvmeMask = RT_BIT_64(12);
+    if (   fIsAmd
+        && (ASMRdMsr(idEfer) & fSvmeMask))
+           return true;
+
+    return false;
+}
+#endif
+
+
 /**
  * Initialize module.
  *
@@ -363,6 +381,10 @@ static int __init VBoxDrvLinuxInit(void)
 {
     int       rc;
 
+    /* Check if modue loading was disabled. */
+    if (!vbox_mod_should_load())
+        return -EINVAL;
+
 #if RTLNX_VER_MIN(5,0,0)
     spin_lock_init(&g_supdrvLinuxWrapperModuleSpinlock);
     RTListInit(&g_supdrvLinuxWrapperModuleList);
@@ -394,21 +416,12 @@ static int __init VBoxDrvLinuxInit(void)
         rc = RTR0Init(0);
         if (RT_SUCCESS(rc))
         {
-#if (defined(RT_ARCH_AMD64) && RTLNX_VER_MAX(2,6,23)) || defined(VBOX_WITH_TEXT_MODMEM_HACK)
-# ifdef VBOX_WITH_TEXT_MODMEM_HACK
-            set_memory_x(&g_abExecMemory[0], sizeof(g_abExecMemory) / PAGE_SIZE);
-            set_memory_rw(&g_abExecMemory[0], sizeof(g_abExecMemory) / PAGE_SIZE);
-# endif
-            rc = RTR0MemExecDonate(&g_abExecMemory[0], sizeof(g_abExecMemory));
-            printk(KERN_DEBUG "VBoxDrv: dbg - g_abExecMemory=%p\n", (void *)&g_abExecMemory[0]);
-#endif
             Log(("VBoxDrv::ModuleInit\n"));
 
             /*
              * Initialize the device extension.
              */
-            if (RT_SUCCESS(rc))
-                rc = supdrvInitDevExt(&g_DevExt, sizeof(SUPDRVSESSION));
+            rc = supdrvInitDevExt(&g_DevExt, sizeof(SUPDRVSESSION));
             if (RT_SUCCESS(rc))
             {
 #ifdef VBOX_WITH_SUSPEND_NOTIFICATION
@@ -427,8 +440,6 @@ static int __init VBoxDrvLinuxInit(void)
                         if (rc2)
                             printk(KERN_WARNING "vboxdrv: failed to register module notifier! rc2=%d\n", rc2);
 #endif
-
-
                         printk(KERN_INFO "vboxdrv: TSC mode is %s, tentative frequency %llu Hz\n",
                                SUPGetGIPModeName(g_DevExt.pGip), g_DevExt.pGip->u64CpuHz);
                         LogFlow(("VBoxDrv::ModuleInit returning %#x\n", rc));
@@ -527,9 +538,9 @@ static int vboxdrvLinuxCreateCommon(struct inode *pInode, struct file *pFilp, bo
      * Only root is allowed to access the unrestricted device, enforce it!
      */
     if (   fUnrestricted
-        && vboxdrvLinuxEuidInNs() != 0 /* root */ )
+        && vboxdrvLinuxEuid() != 0 /* root */ )
     {
-        Log(("VBoxDrvLinuxCreate: euid=%d, expected 0 (root)\n", vboxdrvLinuxEuidInNs()));
+        Log(("VBoxDrvLinuxCreate: euid=%d, expected 0 (root)\n", vboxdrvLinuxEuid()));
         return -EPERM;
     }
 #endif /* VBOX_WITH_HARDENING */
@@ -1022,9 +1033,10 @@ SUPR0DECL(int) SUPDrvLinuxLdrDeregisterWrappedModule(PCSUPLDRWRAPPEDMODULE pWrap
 EXPORT_SYMBOL(SUPDrvLinuxLdrDeregisterWrappedModule);
 
 
+#if defined(RT_ARCH_AMD64) || defined(RT_ARCH_X86)
 RTCCUINTREG VBOXCALL supdrvOSChangeCR4(RTCCUINTREG fOrMask, RTCCUINTREG fAndMask)
 {
-#if RTLNX_VER_MIN(5,8,0)
+# if RTLNX_VER_MIN(5,8,0)
     unsigned long fSavedFlags;
     local_irq_save(fSavedFlags);
     RTCCUINTREG const uOld = cr4_read_shadow();
@@ -1032,25 +1044,26 @@ RTCCUINTREG VBOXCALL supdrvOSChangeCR4(RTCCUINTREG fOrMask, RTCCUINTREG fAndMask
     AssertMsg(cr4_read_shadow() == ((uOld & fAndMask) | fOrMask),
               ("fOrMask=%#RTreg fAndMask=%#RTreg uOld=%#RTreg; new cr4=%#llx\n", fOrMask, fAndMask, uOld, cr4_read_shadow()));
     local_irq_restore(fSavedFlags);
-#else
-# if RTLNX_VER_MIN(3,20,0)
-    RTCCUINTREG const uOld = this_cpu_read(cpu_tlbstate.cr4);
 # else
+#  if RTLNX_VER_MIN(3,20,0)
+    RTCCUINTREG const uOld = this_cpu_read(cpu_tlbstate.cr4);
+#  else
     RTCCUINTREG const uOld = ASMGetCR4();
-# endif
+#  endif
     RTCCUINTREG const uNew = (uOld & fAndMask) | fOrMask;
     if (uNew != uOld)
     {
-# if RTLNX_VER_MIN(3,20,0)
+#  if RTLNX_VER_MIN(3,20,0)
         this_cpu_write(cpu_tlbstate.cr4, uNew);
         __write_cr4(uNew);
-# else
+#  else
         ASMSetCR4(uNew);
-# endif
+#  endif
     }
-#endif
+# endif
     return uOld;
 }
+#endif /* RT_ARCH_AMD64 || RT_ARCH_X86 */
 
 
 void VBOXCALL supdrvOSCleanupSession(PSUPDRVDEVEXT pDevExt, PSUPDRVSESSION pSession)
@@ -1610,7 +1623,7 @@ static DECLCALLBACK(void) supdrvLnxMsrProberModifyOnCpu(RTCPUID idCpu, void *pvU
     if (!fFaster)
         ASMWriteBackAndInvalidateCaches();
 
-    rcBefore = rdmsrl_safe(uMsr, &uBefore);
+    rcBefore = VBOX_RDMSR_SAFE(uMsr, &uBefore);
     if (rcBefore >= 0)
     {
         register uint64_t uRestore = uBefore;
@@ -1619,7 +1632,7 @@ static DECLCALLBACK(void) supdrvLnxMsrProberModifyOnCpu(RTCPUID idCpu, void *pvU
         uWritten |= pReq->u.In.uArgs.Modify.fOrMask;
 
         rcWrite   = wrmsr_safe(uMsr, RT_LODWORD(uWritten), RT_HIDWORD(uWritten));
-        rcAfter   = rdmsrl_safe(uMsr, &uAfter);
+        rcAfter   = VBOX_RDMSR_SAFE(uMsr, &uAfter);
         rcRestore = wrmsr_safe(uMsr, RT_LODWORD(uRestore), RT_HIDWORD(uRestore));
 
         if (!fFaster)
@@ -1665,6 +1678,40 @@ int VBOXCALL    supdrvOSMsrProberModify(RTCPUID idCpu, PSUPMSRPROBER pReq)
 #endif /* SUPDRV_WITH_MSR_PROBER */
 
 
+/**
+ * @copydoc SUPR0EnableHwvirt
+ */
+int VBOXCALL supdrvOSEnableHwvirt(bool fEnable)
+{
+#ifdef SUPDRV_LINUX_HAS_KVM_HWVIRT_API
+    if (fEnable)
+    {
+        if (is_kvm_hwvirt_mod_likely_loaded())
+        {
+            /* kvm_enable_virtualization() is guarded by kvm_usage_count reference counter inside a mutex. */
+            int const rc = kvm_enable_virtualization();
+            if (!rc)
+            {
+                g_fUsingKvmHwvirtApi = true;
+                printk(KERN_INFO "vboxdrv: Using KVM hardware-virtualization API\n");
+                return VINF_SUCCESS;
+            }
+            printk(KERN_ERR "vboxdrv: Failed to enable the KVM kernel API for acquiring VMX/SVM. rc=%d\n", rc);
+        }
+        return VERR_NOT_AVAILABLE;
+    }
+
+    if (g_fUsingKvmHwvirtApi)
+    {
+        kvm_disable_virtualization();
+        g_fUsingKvmHwvirtApi = false;
+    }
+    return VINF_SUCCESS;
+#endif
+    return VERR_NOT_SUPPORTED;
+}
+
+
 /**
  * Converts a supdrv error code to an linux error code.
  *
@@ -1723,12 +1770,25 @@ SUPR0_EXPORT_SYMBOL(SUPR0PrintfV);
 SUPR0DECL(uint32_t) SUPR0GetKernelFeatures(void)
 {
     uint32_t fFlags = 0;
+    /* Note! bird 2023-10-20: Apparently, with CONFIG_PAX_KERNEXEC these days,
+             not only is the regular GDT read-only, but the one returned by
+             get_current_gdt_rw() is also read-only despite the name.
+
+             We don't know exactly when this started, or if it was always like
+             this, but getting hold of the relevant patches isn't all that
+             straight forward any longer it seems (which is weird for linux
+             patches), so, we've just enabled slow-mode for all PAX_KERNEXEC
+             kernels regardless of kernel version.
+
+             Looking at grsecurity patch for 4.9.9, it looks like the writable
+             GDT stuff never worked with PaX/grsec.
+             */
 #ifdef CONFIG_PAX_KERNEXEC
     fFlags |= SUPKERNELFEATURES_GDT_READ_ONLY;
-#endif
-#if RTLNX_VER_MIN(4,12,0)
+#elif RTLNX_VER_MIN(4,12,0)
     fFlags |= SUPKERNELFEATURES_GDT_NEED_WRITABLE;
 #endif
+
 #if defined(VBOX_STRICT) || defined(VBOX_WITH_EFLAGS_AC_SET_IN_VBOXDRV)
     fFlags |= SUPKERNELFEATURES_SMAP;
 #elif defined(CONFIG_X86_SMAP)
@@ -1740,10 +1800,12 @@ SUPR0DECL(uint32_t) SUPR0GetKernelFeatures(void)
 SUPR0_EXPORT_SYMBOL(SUPR0GetKernelFeatures);
 
 
+#if defined(RT_ARCH_AMD64) || defined(RT_ARCH_X86) /* not available on arm64 */
+
 SUPR0DECL(bool) SUPR0FpuBegin(bool fCtxHook)
 {
     RT_NOREF(fCtxHook);
-#if RTLNX_VER_MIN(4,19,0) /* Going back to 4.19.0 for better coverage, we
+# if RTLNX_VER_MIN(4,19,0) /* Going back to 4.19.0 for better coverage, we
                              probably only need 5.17.7+ in the end. */
     /*
      * HACK ALERT!
@@ -1766,17 +1828,24 @@ SUPR0DECL(bool) SUPR0FpuBegin(bool fCtxHook)
      */
     Assert(fCtxHook || !RTThreadPreemptIsEnabled(NIL_RTTHREAD));
     kernel_fpu_begin();
-# if 0 /* Always do it for now for better test coverage. */
+#  if RTLNX_VER_MIN(6,15,0) /* fpregs_unlock may do more than just preempt_enable, so only when necessary now. */
     if (fCtxHook)
-# endif
+#  endif
+    {
+#  if RTLNX_VER_MIN(6,15,0)
+        if (!irqs_disabled())
+            fpregs_unlock();
+#  else
         preempt_enable();
+#  endif
+    }
     return false; /** @todo Not sure if we have license to use any extended state, or
                    *        if we're limited to the SSE & x87 FPU. If it's the former,
                    *        we should return @a true and the caller can skip
                    *        saving+restoring the host state and save some time. */
-#else
+# else
     return false;
-#endif
+# endif
 }
 SUPR0_EXPORT_SYMBOL(SUPR0FpuBegin);
 
@@ -1784,28 +1853,39 @@ SUPR0_EXPORT_SYMBOL(SUPR0FpuBegin);
 SUPR0DECL(void) SUPR0FpuEnd(bool fCtxHook)
 {
     RT_NOREF(fCtxHook);
-#if RTLNX_VER_MIN(4,19,0)
+# if RTLNX_VER_MIN(4,19,0)
     /* HACK ALERT! See SUPR0FpuBegin for an explanation of this. */
     Assert(!RTThreadPreemptIsEnabled(NIL_RTTHREAD));
-# if 0 /* Always do it for now for better test coverage. */
+#  if RTLNX_VER_MIN(6,15,0) /* fpregs_unlock may do more than just preempt_enable, so only when necessary now. */
     if (fCtxHook)
-# endif
+#  endif
+    {
+#  if RTLNX_VER_MIN(6,15,0)
+        if (!irqs_disabled())
+            fpregs_lock();
+#  else
         preempt_disable();
+#  endif
+    }
     kernel_fpu_end();
-#endif
+# endif
 }
 SUPR0_EXPORT_SYMBOL(SUPR0FpuEnd);
 
+#endif /* RT_ARCH_AMD64 || RT_ARCH_X86 */
 
+
+#if defined(RT_ARCH_AMD64) || defined(RT_ARCH_X86)
 int VBOXCALL    supdrvOSGetCurrentGdtRw(RTHCUINTPTR *pGdtRw)
 {
-#if RTLNX_VER_MIN(4,12,0)
+# if RTLNX_VER_MIN(4,12,0) && !defined(CONFIG_PAX_KERNEXEC)
     *pGdtRw = (RTHCUINTPTR)get_current_gdt_rw();
     return VINF_SUCCESS;
-#else
+# else
     return VERR_NOT_IMPLEMENTED;
-#endif
+# endif
 }
+#endif
 
 
 module_init(VBoxDrvLinuxInit);
